import os
import io
import sys
from typing import NamedTuple

#Global
#Get current directory
workingDir = os.path.dirname(os.path.realpath(__file__))

#Store per file binary data
simMeshes = []  #We can have multiple sim nodes per ycl
isWorkingFile = False
all_bytes = b''

#Store each node as an object to retain properties
class simNode(NamedTuple):
    index: int
    offset: int
    size: int
    binary: bytes


def getDirFiles():
    global workingDir
    dir_list = os.listdir(workingDir)

    #Return all files as list
    return dir_list

def getAllYclFilesFromDir( files ):
    yclPaths = []
    for path in files:
        if path.endswith(".ycl"):
            yclPaths.append( path )

    if len(yclPaths) == 0: 
        print( "Found no YCL files in your directory.")
        return False
    else: 
        print( "YCL Files: " ,  yclPaths )
        return yclPaths

def findBufferMagic( file , blockSig , bufferPos ):
    #Catch if we can't find buffer
    bufferSig , bufferStart , blockLength = None, None , None
    fileSize = os.path.getsize(file.name)
    file.seek( bufferPos )

    #Recurse for 0xblockSig buffer
    while( bufferSig != blockSig and ( bufferPos< fileSize ) ):
        bufferPos = file.tell()
        bufferSig = int.from_bytes( file.read(4) , "little" )
        bufferStart = int.from_bytes( file.read(4) , "little" )
        blockLength = int.from_bytes( file.read(4) , "little" )

        #Use 0x5 as sim entry indicator
        if(bufferSig == 0x5):
            node = simNode(index = len(simMeshes), offset=bufferPos, size=blockLength, binary=None)
            simMeshes.append(node)

        if(bufferSig==0): blockLength = 0x20
        if(bufferSig==0 and bufferPos!= 0): break
        if(bufferSig!=blockSig): file.seek( bufferPos + blockLength )

    return bufferPos , bufferStart , (bufferSig==blockSig) , blockLength


def formatDataTableBinary( dataTable ):
    fmtBuffer = io.BytesIO(b'')

    for item in dataTable:
        fmtBuffer.seek( 0 , 2 ); 
        fmtBuffer.write( int( item[1] ).to_bytes( 4 , "little" ) ) #floatX
        fmtBuffer.seek( 0 , 2 ); 
        fmtBuffer.write( int( item[2] ).to_bytes( 4 , "little" ) ) #floatY
        fmtBuffer.seek( 0 , 2 ); 
        fmtBuffer.write( int( item[3] ).to_bytes( 4 , "little" ) ) #floatZ
        fmtBuffer.seek( 0 , 2 ); 
        fmtBuffer.write( int( item[0] ).to_bytes( 4 , "little" ) ) #index

    return fmtBuffer.getvalue()


def getNewFileString( string ):
    newYCLName = string.split("\\")[-1].split(".")[0]
    path =  string.split("\\"); path.pop()
    return ( '\\'.join(path) + "\\" + ( newYCLName + "_reformatted.ycl" ) )

#This will preserve any previous reformats
def cacheFileBinaries(path):
    #read all old file data
    with open(path, "rb") as file:
        file.seek(0)
        all_bytes = file.read()
    return all_bytes

def createReformatYCLFile( newTableBinary ,items , bufferPos , bufferSize, file ):
    #Create new file name string and path
    #We will ignore if we are already working on our reformat file
    global all_bytes
    path =  getNewFileString( file.name )
    canWrite = True

    #Check file access
    if( os.path.exists(path) ):
        canWrite = os.access(path, os.W_OK)
    else: 
        with open(path, "w") as f:
            f.write(''); f.close()

    if canWrite:
        #reset out file
        with open(path, "w"): pass

        #write binaries
        with open(path, "rb+") as f:
            #first write all old contents
            f.write(all_bytes); f.seek(bufferPos)
            f.write(b"\x00" * bufferSize); f.seek(bufferPos);
            
            #then we add our new buffer
            f.write( int( 0x1D ).to_bytes( 4 , "little" ) ) #bufferMagic
            f.write( int( 0x10 ).to_bytes( 4 , "little" ) ) #bufferStart
            f.write( int( bufferSize ).to_bytes( 4 , "little" ) ) #bufferSize
            f.write( int( items ).to_bytes( 4 , "little" ) ) #total items
            f.write(newTableBinary); f.close()

        if (not isWorkingFile): print("Wrote to file: " , path )
        #Store binaries to preserve changes
        all_bytes = cacheFileBinaries(path)
    else:
        print("Could not write to file: ", path)


    return 

def evaluateBuffer(file,bufferPos):
    #Collect basic pointer data
    file.seek(bufferPos) 
    bufferSig = int.from_bytes( file.read(4) , "little" )
    bufferStart = int.from_bytes( file.read(4) , "little" )
    bufferSize = int.from_bytes( file.read(4) , "little" )
    

    #Interpret Actual Data buffer
    is23YCL = False
    itemTotal = int.from_bytes( file.read(4) , "little" )
    dataTable = []

    #Loop and gather
    for i in range(itemTotal):
        index = int.from_bytes( file.read(4) , "little" )
        float_x = int.from_bytes( file.read(4) , "little" )
        float_y = int.from_bytes( file.read(4) , "little" )
        float_z = int.from_bytes( file.read(4) , "little" )
        nullDWORD = int.from_bytes( file.read(4) , "little" )

        #Store item data
        dataTable.append( [index,float_x,float_y,float_z] )

        #We use this value to check if this is formatted
        if (nullDWORD != 0 and float_x != 0): is23YCL = True; break

    fileName = file.name.split("\\")[-1]

    if( is23YCL ):
        print( fileName + ": No changes made. This file might be formatted already.")      
    else:
        if (not isWorkingFile): print( fileName + ": File requires reformat.")
        newTableBinary = formatDataTableBinary(dataTable)
        createReformatYCLFile( newTableBinary , len(dataTable) , bufferPos , bufferSize , file )

    return is23YCL


def assessYCL(file):
    #Creates full file path
    fileLog = "(" +  file + ")" + " "
    filePath = workingDir + "\\" + file
    hasBuffer = False

    global all_bytes
    global isWorkingFile

    with open(filePath, 'rb') as file:
        print("\n")

        #First Collect All Sim Nodes
        bufferPos = 0; bufferLength = 0
        while( bufferPos < os.path.getsize(file.name)  ):
            bufferPos,bufferStart,bufferSig,bufferLength = findBufferMagic( file , 0x5 , bufferPos+bufferLength )
        
        print( fileLog + " Nodes Found: " , len(simMeshes) )

        #Seek to node entry end
        bufferPos = ( simMeshes[-1].offset + simMeshes[-1].size )
        #Store binaries to preserve changes
        all_bytes = cacheFileBinaries(filePath)

        #iterate through every node and search for our faulty buffers
        for i in range( len(simMeshes) ):

            #Seek 0x1B buffer
            bufferPos,bufferStart,bufferSig,bufferLength = findBufferMagic( file , 0x1B , bufferPos )

            #Check valid buffer
            if (bufferSig):
                #Seek 0x1C buffer
                bufferStart += bufferPos
                bufferPos,bufferStart,bufferSig,bufferLength = findBufferMagic( file , 0x1D , bufferStart )

                #Evaluate if we succesfully found our block
                if (bufferSig):
                    print( fileLog  + "Found 0x1D Block @ ", bufferPos )
                    hasBuffer = True
                    evaluateBuffer(file,bufferPos)

                    #Update our file to our working reformat file to continue inspecting
                    if (not isWorkingFile): isWorkingFile = True
    
    if(not hasBuffer): 
        print( file.name , ": No changes made. This file might be formatted already.")      

    return


def readYCLFiles(yclFiles):
    global simMeshes
    global isWorkingFile
    global all_bytes

    for path in yclFiles:
        #Clear all old globals per file
        simMeshes = []
        isWorkingFile = False
        all_bytes = b''

        assessYCL(path)
    return


if __name__ == "__main__":

    #Collect all files
    yclFiles = getAllYclFilesFromDir( getDirFiles() )

    if(yclFiles):
        readYCLFiles(yclFiles)

    input("Press any key to exit...")